import requests

# TODO: add params/payload and response validation


def get_request(url, params=None, headers=None, timeout=10):
    response = requests.get(
        url, params=params, headers=headers, timeout=timeout, verify=False
    ).json()
    return response


def post_request(url, payload=None, headers=None, timeout=10):
    response = requests.post(
        url, json=payload, headers=headers, timeout=timeout, verify=False
    ).json()
    return response


def typename(obj):
    return type(obj).__name__



from .authmixin import HEADER_JSON, AuthError, AuthMixin
from .httpclient import HttpClient


class SciCatRestAPI(HttpClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.login_path = self._login_path or "https://dacat.psi.ch/auth/msad"

    def authenticate(self, username, password):
        auth_payload = {"username": username, "password": password}
        res = self._login(auth_payload, HEADER_JSON)
        try:
            token = res["id"]
        except KeyError as e:
            raise SciCatAuthError(res) from e
        else:
            return token


class SciCat:
    def __init__(self, *args, **kwargs):
        self.http_client = SciCatRestAPI(*args, **kwargs)

    @property
    def proposals(self):
        url = self.http_client.address + "/proposals"
        return self.http_client.get_request(url, headers=HEADER_JSON)


class SciCatAuthError(AuthError):
    pass



import getpass
from abc import ABC, abstractmethod
from json.decoder import JSONDecodeError

from .config import Config
from .utils import typename

HEADER_JSON = {"Content-type": "application/json", "Accept": "application/json"}


class AuthMixin(ABC):
    def __init__(self, address, options=None):
        self.address = address.rstrip("/")
        tn = typename(self).lower()
        try:
            self.config = Config(f".{tn}-tokens")
        except JSONDecodeError:
            self.config = {}
        if not options:
            options = {}
        self._token = options.get("token")
        self._username = options.get("username")
        self._password = options.get("password")
        self._login_path = options.get("login_path")

    def __repr__(self):
        tn = typename(self)
        return f"{tn} @ {self.address}"

    @abstractmethod
    def authenticate(self, username, password):
        raise NotImplementedError

    @property
    def token(self):
        return self._retrieve_token()

    def _retrieve_token(self):
        username = self._username or getpass.getuser()
        token = self._token
        if token is None:
            try:
                token = self.config[username]
            except KeyError:
                tn = typename(self)
                password = self._password or getpass.getpass(
                    prompt=f"{tn} password for {username}: "
                )
                token = self.authenticate(username, password)
        self.config[username] = self._token = token
        return token


class AuthError(Exception):
    pass



import functools
import logging
from typing import get_type_hints

logger = logging.getLogger(__name__)


def scilog_typechecked(func):
    @functools.wraps(func)
    def typechecked_call(obj, *args, **kwargs):
        type_hints = get_type_hints(func)
        del type_hints["return"]
        for arg, dtype in zip(args, type_hints.values()):
            arg_type = type(arg)
            if dtype != arg_type and arg is not None:
                raise TypeError(
                    f"{func} expected to receive input of type {dtype.__name__} but received {arg_type.__name__}"
                )
            if len(func.__closure__) > 0:
                property_name = func.__closure__[0].cell_contents.strip("_")
                if obj._deprecated.get(property_name):
                    logger.warning(
                        f"{property_name} is deprecated by {obj._deprecated_by[property_name]}"
                    )
        return func(obj, *args, **kwargs)

    return typechecked_call


def property_maker(name, dtype):
    storage_name = "_" + name

    @property
    def prop(self) -> dtype:
        return getattr(self, storage_name)

    @prop.setter
    @scilog_typechecked
    def prop(self, value: dtype) -> None:
        setattr(self, storage_name, value)

    return prop


class Snippet:
    def __init__(self, snippetType="snippet"):
        self._properties = []
        self.init_properties(snippetType=str)
        self.snippetType = snippetType

    def init_properties(self, **kwargs):
        for name, dtype in kwargs.items():
            storage_name = "_" + name
            cls = type(self)
            setattr(cls, storage_name, None)
            setattr(cls, name, property_maker(name, dtype))
            self._properties.append(name)

    def to_dict(self, include_none=True, include_deprecated=True):
        properties = (
            self._properties
            if include_deprecated
            else list(set(self._properties).difference(self._deprecated.keys()))
        )
        if include_none:
            return {key: getattr(self, key) for key in properties}
        return {key: getattr(self, key) for key in properties if getattr(self, key) is not None}

    def import_dict(self, properties):
        for name, value in properties.items():
            setattr(self, name, value)

    @classmethod
    def from_dict(cls, properties):
        new = cls()
        new.import_dict(properties)
        return new

    def __str__(self):
        return f"{self.__class__.__name__}: {self.to_dict(include_none=False)}"

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.to_dict(include_none=False)}"

    @classmethod
    def from_http_response(cls, response):
        if isinstance(response, list):
            return [cls.from_dict(resp) for resp in response]
        else:
            return cls.from_dict(response)


class Basesnippet(Snippet):
    _deprecated = {"ownerGroup": str, "accessGroups": list}
    _deprecated_by = {"ownerGroup": "ACLS", "accessGroups": "ACLS"}

    def __init__(self, snippetType="basesnippet"):
        super().__init__(snippetType=snippetType)
        self.init_properties(
            id=str,
            parentId=str,
            createACL=list,
            readACL=list,
            updateACL=list,
            deleteACL=list,
            shareACL=list,
            adminACL=list,
            isPrivate=bool,
            createdAt=str,
            createdBy=str,
            updatedAt=str,
            updateBy=str,
            subsnippets=list,
            tags=list,
            dashboardName=str,
            files=list,
            location=str,
            defaultOrder=int,
            linkType=str,
            versionable=bool,
            deleted=bool,
            **self._deprecated,
        )


class Location(Basesnippet):
    def __init__(self):
        super().__init__(snippetType="location")
        self.init_properties(name=str, location=str)


class Logbook(Basesnippet):
    def __init__(self):
        super().__init__(snippetType="logbook")
        self.init_properties(name=str, description=str, thumbnail=str, location=str)


class Paragraph(Basesnippet):
    def __init__(self):
        super().__init__(snippetType="paragraph")
        self.init_properties(textcontent=str, isMessage=str)


class Filesnippet(Basesnippet):
    def __init__(self):
        super().__init__(snippetType="image")  # TODO make that type "file" eventually ?
        self.init_properties(fileExtension=str, accessHash=str)


def snippet_factory(response):
    if isinstance(response, list):
        return [get_snippet_response(resp) for resp in response]
    return get_snippet_response(response)


def get_snippet_response(response):
    # extract snippetType from response
    snippetType = response["snippetType"]
    if snippetType == "location":
        return Location.from_http_response(response)
    if snippetType == "logbook":
        return Logbook.from_http_response(response)
    if snippetType == "paragraph":
        return Paragraph.from_http_response(response)
    if snippetType == "image":
        return Filesnippet.from_http_response(response)
    if snippetType == "basesnippet":
        return Basesnippet.from_http_response(response)



from __future__ import annotations

import functools
import json
import logging
import os
import uuid
import warnings
from typing import Any, Tuple

from typeguard import typechecked

import scilog.logbook_message as lm
from .authmixin import HEADER_JSON, AuthError
from .httpclient import HttpClient
from .snippet import Basesnippet, Filesnippet, Location, Paragraph, snippet_factory

ACLS = ["createACL", "readACL", "updateACL", "deleteACL", "shareACL", "adminACL"]


logger = logging.getLogger(__name__)


def pinned_to_logbook(logbook_keys, include_none=False):
    def pinned_to_logbook_inner(func):
        @functools.wraps(func)
        def pinned_to_logbook_call(log, *args, **kwargs):
            if not isinstance(log.logbook, Basesnippet):
                warnings.warn("No logbook selected.")
            else:
                logbook = log.logbook.to_dict(include_none=include_none)
                logger.debug(f"====== pinned to logbook: {kwargs}")
                for key in logbook_keys:
                    logger.debug(f"======pinned to logbook: {key}")
                    if key not in kwargs:
                        if key == "parentId":  # mapping from childinstance to parentinstance
                            if "where" not in kwargs:
                                kwargs[key] = log.logbook.id
                            else:
                                kwargs["where"] = {
                                    "and": [kwargs["where"], {"parentId": log.logbook.id}]
                                }
                        else:
                            logger.debug(f"Non parentId condition in pinned_to_logbook: {key}")
                            if logbook.get(key):
                                kwargs[key] = logbook[key]
                logger.debug(f"========= pinned to logbook: resulting kwargs: {kwargs}")
            return func(log, *args, **kwargs)  # real call happens here

        return pinned_to_logbook_call

    return pinned_to_logbook_inner


class SciLogRestAPI(HttpClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SciLogCore:
    IMAGE_TYPES = ["png", "jpg", "jpeg"]

    def __init__(self, high_level_interface) -> None:
        self._hli = high_level_interface

    @property
    def logbook(self):
        return self._hli.logbook

    @logbook.setter
    def logbook(self, val):
        self._hli.logbook = val

    @property
    def http_client(self):
        return self._hli.http_client

    def make_filter(
        self,
        where: dict = None,
        limit: int = 0,
        skip: int = 0,
        fields: dict = None,
        include: dict = None,
        order: list = None,
    ):
        filt = dict()
        if where is not None:
            # TOOD needed ? items = [where.copy()]
            filt["where"] = where  #
        if limit > 0:
            filt["limit"] = limit
        if skip > 0:
            filt["skip"] = skip
        if fields is not None:
            filt["fields"] = fields
        if include is not None:
            filt["include"] = include
        if order is not None:
            filt["order"] = order
        filt = json.dumps(filt)
        return {"filter": filt}

    @staticmethod
    def _replace_json_placeholder(snippet: dict, field: str, data: Any) -> dict:
        if not snippet.get(field):
            return snippet
        if isinstance(snippet[field], list):
            if "default" in snippet[field]:
                data = set(data) | set(group for group in snippet[field] if group)
                snippet[field] = list(data)
            return snippet

        if isinstance(snippet[field], str):
            if snippet[field] == "default":
                snippet[field] = data
            return snippet

        raise ValueError("The used placeholder type is not supported. ")

    # TODO if this is a user function add docstring
    @pinned_to_logbook(["parentId", *ACLS])
    def post_snippet(self, **kwargs):
        url = self.http_client.address + "/basesnippets"
        payload = kwargs
        if payload.get("files"):
            payload = self.upload_files(payload)
        return Paragraph.from_http_response(
            self.http_client.post_request(url, payload=payload, headers=HEADER_JSON)
        )

    # TODO if this is a user function add docstring
    @pinned_to_logbook(["parentId", *ACLS])
    def post_location(self, **kwargs):
        url = self.http_client.address + "/basesnippets"
        payload = kwargs
        if payload.get("files"):
            payload = self.upload_files(payload)
        return Location.from_http_response(
            self.http_client.post_request(url, payload=payload, headers=HEADER_JSON)
        )

    def upload_files(self, payload):
        for file in payload.get("files"):
            if "fileId" in file and file["fileId"] is not None:
                continue
            logger.debug(f"Posting from filepath: {file['filepath']}")
            filesnippet = self._post_filesnippet(
                file["filepath"],
                **{key: val for key, val in payload.items() if key in [*ACLS, "tags"]},
            )
            file["fileId"] = filesnippet.id
            file["accessHash"] = filesnippet.accessHash
            file.pop("filepath")
        return payload

    def _post_filesnippet(self, filepath, **kwargs):
        url = self.http_client.address + "/filesnippet/files"

        file_extension = os.path.splitext(filepath)[-1]
        if not file_extension:
            raise ValueError("filepath must be pointing to a file, not a directory.")
        file_extension = file_extension[1:]

        snippet = Filesnippet()
        snippet.import_dict(kwargs)
        snippet.fileExtension = file_extension
        payload = snippet.to_dict(include_none=False)

        file_descriptor = "image" if file_extension in self.IMAGE_TYPES else "file"
        multipart_form_data = {
            "file": (
                filepath + "." + file_extension,
                open(filepath, "rb"),
                f"{file_descriptor}/{file_extension}",
            ),
            "fields": (None, json.dumps(payload)),
        }
        return Filesnippet.from_http_response(
            self.http_client.post_request(
                url, files=multipart_form_data, headers={"Accept": "application/json"}
            )
        )

    @pinned_to_logbook(ACLS)
    def post_file(self, filepath: str, **kwargs) -> Filesnippet:
        """Upload a file. As they are disconnected from paragraphs and logbooks, there
        is no need to copy the parentId. Instead we simply keep the ACLS.

        Args:
            filepath (str): Path to the file that ought to be uploaded
        Raises:
            FileNotFoundError: Raised if the file specified in filepath does not exist
            ValueError: Raised if the filepath is not pointing to a file
        Returns:
            Filesnippet: Filesnippet containing the metadata of the newly created entry
        """
        if not os.path.exists(filepath):
            raise FileNotFoundError("Specified file does not exist.")

        fsnippet = self._post_filesnippet(filepath, **kwargs)
        # ret = self._file_upload(filepath, fsnippet.id, file_extension)
        return fsnippet

    @pinned_to_logbook(ACLS)
    def append_files_to_snippet(self, snippet: Paragraph, filepaths: list, **kwargs) -> Paragraph:
        """Append files or images to an already existing snippet. Files and images will be
        appended following the order given in 'filepaths'. As the files / images are disconnected
        from paragraphs and logbooks, there is no need to copy the parentId. Instead we simply
        keep the ACLS.

        Args:
            snippet (Paragraph): Snippet to which the files should be appended
            filepaths (list): List of file paths pointing to the files that should be uploaded
        Returns:
            Paragraph: Updated snippet
        """
        for filepath in filepaths:
            fsnippet = self.post_file(filepath)

            if not isinstance(snippet.files, list):
                snippet.files = []

            # if we reach this point, we can assume that filepath has been checked (cf. self.post_file)
            file_info, file_textcontent = self.prepare_file_content(
                filepath=filepath, fsnippet=fsnippet
            )
            snippet.textcontent += file_textcontent
            snippet.files.append(file_info)

        return self.patch_snippet(snippet)

    @staticmethod
    def prepare_file_content(filepath: str, fsnippet: Filesnippet = None) -> Tuple:
        file_extension = filepath.split(".")[-1].lower()
        file_hash = str(uuid.uuid4())
        file_id = fsnippet.id if fsnippet and fsnippet.id else None
        accessHash = fsnippet.accessHash if fsnippet and fsnippet.accessHash else None

        if file_extension in SciLogCore.IMAGE_TYPES:
            textcontent = (
                f'<figure class="image image_resized"><img src="" title="{file_hash}"></figure>'
            )
            file_info = {
                "fileHash": file_hash,
                "filepath": filepath,
                "fileExtension": f"image/{file_extension}",
                "fileId": file_id,
                "accessHash": accessHash,
                "style": {"width": "82.25%", "height": ""},
            }

        else:
            textcontent = f'<p><a class="fileLink" target="_blank" href="file:{file_hash}">{os.path.basename(filepath)}</a></p>'
            file_info = {
                "fileHash": file_hash,
                "filepath": filepath,
                "fileExtension": f"file/{file_extension}",
                "fileId": file_id,
                "accessHash": accessHash,
            }
        return (file_info, textcontent)

    def patch_snippet(self, snippet: Basesnippet, **kwargs) -> Basesnippet:
        """Update (patch) snippet with given snippet.
        Args:
            snippet (Basesnippet): Snippet containing the newly updated fields
        Returns:
            Basesnippet: Updated snippet
        """
        url = self.http_client.address + "/basesnippets/" + snippet.id
        snippet.id = None
        snippet.createdAt = None
        snippet.createdBy = None
        snippet.updatedAt = None
        snippet.updatedBy = None
        snippet.expiresAt = None
        payload = snippet.to_dict(include_none=False)
        return snippet_factory(
            self.http_client.patch_request(url, payload=payload, headers=HEADER_JSON)
        )


class SciLog:
    def __init__(self, *args, **kwargs):
        self.http_client = SciLogRestAPI(*args, **kwargs)
        self.logbook = None
        self.core = SciLogCore(self)

    def select_logbook(self, logbook: Basesnippet):
        self.logbook = logbook

    @pinned_to_logbook(["parentId"])
    def get_snippets(
        self,
        where: dict = None,
        limit: int = 0,
        skip: int = 0,
        fields: dict = None,
        include: dict = None,
        order: list = None,
        **kwargs,
    ):
        url = self.http_client.address + "/basesnippets"
        if not where:
            items = [kwargs.copy()]
            where_keys = {"and": items}
            logger.debug(f"==== where condition defined via keyword arguments:{where_keys}")
            params = self.core.make_filter(
                where=where_keys,
                limit=limit,
                skip=skip,
                fields=fields,
                include=include,
                order=order,
            )
        else:
            logger.debug(f"==== where condition defined directly:{where}")
            params = self.core.make_filter(
                where=where, limit=limit, skip=skip, fields=fields, include=include, order=order
            )
        # TODO dont allow case that both where and kwargs are defined, return error in this case
        return snippet_factory(
            self.http_client.get_request(url, params=params, headers=HEADER_JSON)
        )

    @pinned_to_logbook(["parentId", *ACLS])
    def import_from_dict(self, snippet: dict, **kwargs):
        msg = snippet["textcontent"]
        snippet_dict = {k: v for k, v in snippet.items() if k != "textcontent"}
        self.send_message(msg, **snippet_dict)

    @pinned_to_logbook(["parentId", *ACLS])
    def send_message(self, msg, **kwargs):
        lm_msg = lm.LogbookMessage()
        lm_msg._content.import_dict(kwargs)
        lm_msg.add_text(msg)

        return self.send_logbook_message(lm_msg)

    @typechecked
    def send_logbook_message(self, msg: lm.LogbookMessage) -> None:
        """Upload a new message to SciLog.

        Args:
            msg (lm.LogbookMessage): Previously composed LogbookMessage.

        Example:
            >>> msg = LogbookMessage()
            >>> msg.add_text("Another example text").add_tag(["color"])
            >>> log.send_logbook_message(msg)

        """
        payload = msg._content.to_dict(include_none=False)
        payload["linkType"] = "paragraph"
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.core.post_snippet(**payload)

    def get_logbooks(
        self,
        where: dict = None,
        limit: int = 0,
        skip: int = 0,
        fields: dict = None,
        include: dict = None,
        order: list = None,
        **kwargs,
    ):
        url = self.http_client.address + "/logbooks"
        if not where:
            items = [kwargs.copy()]
            where_keys = {"and": items}
            logger.debug(f"==== where condition defined via keyword arguments:{where_keys}")
            params = self.core.make_filter(
                where=where_keys,
                limit=limit,
                skip=skip,
                fields=fields,
                include=include,
                order=order,
            )
        else:
            logger.debug(f"==== where condition defined directly:{where}")
            params = self.core.make_filter(
                where=where, limit=limit, skip=skip, fields=fields, include=include, order=order
            )
        return snippet_factory(
            self.http_client.get_request(url, params=params, headers=HEADER_JSON)
        )

    # TODO: review and potentially move the whole core logic here
    def post_snippet(self, **kwargs):
        return self.core.post_snippet(**kwargs)


class SciLogAuthError(AuthError):
    pass
